import asyncio
from contextlib import asynccontextmanager, contextmanager
from typing import Any, Optional, Type

from sentry_sdk import Scope
from tornado.iostream import IOStream, StreamClosedError

from cats.errors import HandshakeError, ProtocolError
from cats.identity import Identity
from cats.server.action import BaseAction, Input
from cats.server.app import Application
from cats.server.handlers import Handler
from cats.types import BytesAnyGen, T_Headers


class Connection:
    MAX_PLAIN_DATA_SIZE: int
    STREAM_ERRORS = (asyncio.TimeoutError, asyncio.CancelledError, asyncio.InvalidStateError, StreamClosedError)
    IGNORE_ERRORS = (*STREAM_ERRORS, HandshakeError, ProtocolError, KeyboardInterrupt)

    __slots__ = (
        '_closed', 'stream', 'address', 'api_version', '_app', 'scope', 'download_speed',
        '_identity', '_credentials', 'loop', 'input_deq', '_idle_timer', 'message_pool', 'is_sending',
        'debug', 'recv_fut',
    )

    def __init__(self, stream: IOStream, address: tuple[str, int], api_version: int, app: Application,
                 debug: bool = False):
        self._closed: bool = False
        self.stream: IOStream = stream
        self.address: tuple[str, int] = address
        self.api_version: int = api_version
        self._app: Application = app
        self.scope: Scope = Scope()
        self._identity: Optional[Identity] = None
        self._credentials: Any = None
        self.loop: asyncio.BaseEventLoop = asyncio.get_event_loop()
        self.input_deq: dict[int, Input] = {}
        self._idle_timer: Optional[asyncio.Future] = None
        self.message_pool: list[int] = []
        self.is_sending: bool = False
        self.download_speed: int = 0
        self.debug = debug
        self.recv_fut: Optional[asyncio.Future] = None

    @property
    def host(self) -> str: ...

    @property
    def port(self) -> int: ...

    @property
    def is_open(self) -> bool: ...

    @property
    def app(self) -> Application: ...

    async def init(self) -> None: ...

    async def start(self, ping: bool = True) -> None: ...

    @contextmanager
    def preserve_message_id(self, message_id: int): ...

    async def recv(self) -> BaseAction: ...

    async def tick(self) -> None: ...

    def on_tick_done(self, task: asyncio.Task) -> None: ...

    def dispatch(self, handler_id: int) -> Type[Handler]: ...

    async def ping(self) -> None: ...

    async def set_download_speed(self, speed: int = 0): ...

    async def send(self, handler_id: int, data: Any, message_id: int = None, compression: int = None, *,
                   headers: T_Headers = None, status: int = 200) -> None: ...

    async def send_stream(self, handler_id: int, data: BytesAnyGen, data_type: int,
                          message_id: int = None, compression: int = None, *,
                          headers: T_Headers = None, status: int = 200) -> None: ...

    def attach_to_channel(self, channel: str): ...

    def detach_from_channel(self, channel: str): ...

    @property
    def identity(self) -> Optional[Identity]: ...

    @property
    def credentials(self) -> Optional[Any]: ...

    @property
    def identity_scope_user(self): ...

    def signed_in(self) -> bool: ...

    def sign_in(self, identity: Identity, credentials: Any = None): ...

    def sign_out(self): ...

    def close(self, exc: BaseException = None) -> None: ...

    def __str__(self) -> str: ...

    class ProtocolError(ValueError, IOError):
        pass

    def reset_idle_timer(self) -> None: ...

    def _get_free_message_id(self) -> int: ...

    @asynccontextmanager
    async def lock_write(self): ...



import asyncio
import tempfile
import warnings
from importlib import import_module
from logging import getLogger
from pathlib import Path
from time import time
from typing import Union

import math

__all__ = [
    'Delay',
    'require',
    'tmp_file',
    'int2hex',
    'bytes2hex',
    'format_amount',
    'enable_stream_debug',
]

logging = getLogger('CATS.utils')


class Delay:
    def __init__(self, speed: int):
        self.speed = speed

    def __enter__(self):
        self.start = time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        ...

    async def __call__(self):
        if not self.speed:
            return

        to_sleep = max(0.0, min(1.0, 1.0 - time() + self.start))
        self.start = time()
        if to_sleep > 0:
            print(f' SLEEP {to_sleep:0.4f}'.center(64, '='))
            await asyncio.sleep(to_sleep)


def tmp_file(**kwargs) -> Path:
    kwargs['delete'] = kwargs.get('delete', False)
    return Path(tempfile.NamedTemporaryFile(**kwargs).name)


def require(dotted_path: str, /, *, strict: bool = True):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        try:
            module_path, class_name = dotted_path.rsplit('.', 1)
        except ValueError as err:
            raise ImportError(f"{dotted_path} doesn't look like a module path") from err

        module = import_module(module_path)

        try:
            return getattr(module, class_name)
        except AttributeError as err:
            raise ImportError(f'Module "{module_path}" does not define a "{class_name}" attribute/class') from err
    except ImportError as err:
        logging.error(f'Failed to import {dotted_path}')
        if strict:
            raise err
        return None


def int2hex(number: int, chars: int = 2) -> str:
    a = hex(number)[2:].upper()
    chars = max(len(a) // 2, chars)
    return ('00' * chars + a)[-chars * 2:]


def bytes2hex(buffer: Union[bytes, bytearray, memoryview], *, separator: str = ' ', prefix: bool = False) -> str:
    """
    Converts byte-like to HEX string

    :param buffer: what to encode
    :param separator: separator string
    :param prefix: use 0x prefix
    :return: HEX style string
    :raise TypeError:
    """
    if not isinstance(buffer, (bytes, bytearray, memoryview)):
        raise TypeError(f'Invalid buffer type = {type(buffer)}')

    hexadecimal = buffer.hex().upper()
    parts = (hexadecimal[i: i + 2] for i in range(0, len(hexadecimal), 2))
    if prefix:
        parts = ('0x' + part for part in parts)
    return separator.join(parts)


def format_amount(num: int, *, base: int = 1024, prefix: str = '', suffix: str = 'B') -> str:
    """
    Humanize any amount to string
    format_amount(0) == "0B"
    format_amount(1750) == "1.7KB"
    format_amount(1234567890) == "1.1GB"
    format_amount(2000, base=10, suffix='r') == "2Gr"

    :param num: numeric amount
    :param base: how much does the step takes (default 1024 as for bytes)
    :param prefix: type prefix (default '')
    :param suffix: type suffix (default B as for bytes)
    :return: Formatted amount string
    :raise TypeError:
    """
    if not num:
        return f'0{suffix}'

    if not isinstance(num, int):
        raise TypeError(f'Invalid num type = {type(num)}')

    if not isinstance(base, int):
        raise TypeError(f'Invalid base type = {type(base)}')

    sign = '-' if num < 0 else ''
    num = abs(num)

    magnitude = int(math.floor(math.log(num, base)))
    val = float(num / math.pow(base, magnitude))

    if magnitude > 7:
        val = f'{val:.0f}' if val.is_integer() else f'{val:.1f}'
        return f'{sign}{val}{"Y"}{prefix}{suffix}'

    val = f'{val:.0f}' if val.is_integer() else f'{val:3.1f}'
    return f'{sign}{val}{["", "K", "M", "G", "T", "P", "E", "Z"][magnitude]}{prefix}{suffix}'


def enable_stream_debug():
    warnings.warn("enable_stream_debug will be deprecated in version 4.0. "
                  "Use Server(debug=True) or Connection(debug=True) instead", PendingDeprecationWarning)
    from tornado.iostream import IOStream
    rb = IOStream.read_bytes
    ru = IOStream.read_until
    wr = IOStream.write

    async def read_bytes(self: IOStream, num_bytes, partial: bool = False):
        chunk = await rb(self, num_bytes, partial=partial)
        peer = self.socket.getpeername() if self.socket else ('0.0.0.0', 0)
        logging.info(f'[RECV {peer}] {bytes2hex(chunk)}')
        return chunk

    async def read_until(self: IOStream, delimiter: bytes, max_bytes: int = None):
        chunk = await ru(self, delimiter, max_bytes=max_bytes)
        peer = self.socket.getpeername() if self.socket else ('0.0.0.0', 0)
        logging.info(f'[RECV {peer}] {bytes2hex(chunk)}')
        return chunk

    async def write(self: IOStream, data):
        peer = self.socket.getpeername() if self.socket else ('0.0.0.0', 0)
        logging.info(f'[SEND {peer}] {bytes2hex(data)}')
        return await wr(self, data)

    IOStream.read_until = read_until
    IOStream.read_bytes = read_bytes
    IOStream.write = write



import gzip
import os
import shutil
from pathlib import Path
from typing import Union

__all__ = [
    'BaseCompressor',
    'DummyCompressor',
    'GzipCompressor',
    'Compressor',
]


class BaseCompressor:
    type_id: int
    type_name: str

    @classmethod
    async def compress(cls, data: bytes) -> bytes:
        raise NotImplementedError

    @classmethod
    async def decompress(cls, data: bytes) -> bytes:
        raise NotImplementedError

    @classmethod
    async def compress_file(cls, src: Path, dst: Path) -> None:
        raise NotImplementedError

    @classmethod
    async def decompress_file(cls, src: Path, dst: Path) -> None:
        raise NotImplementedError


class DummyCompressor(BaseCompressor):
    type_id = 0x00
    type_name = 'dummy'

    @classmethod
    async def compress(cls, data: bytes) -> bytes:
        return data

    @classmethod
    async def decompress(cls, data: bytes) -> bytes:
        return data

    @classmethod
    async def compress_file(cls, src: Path, dst: Path) -> None:
        shutil.copy(src.resolve().as_posix(), dst.resolve().as_posix())

    @classmethod
    async def decompress_file(cls, src: Path, dst: Path) -> None:
        shutil.copy(src.resolve().as_posix(), dst.resolve().as_posix())


class GzipCompressor(BaseCompressor):
    type_id = 0x01
    type_name = 'GZip'

    @classmethod
    async def compress(cls, data: bytes) -> bytes:
        return gzip.compress(data, compresslevel=9)

    @classmethod
    async def decompress(cls, data: bytes) -> bytes:
        return gzip.decompress(data)

    @classmethod
    async def compress_file(cls, src: Path, dst: Path) -> None:
        with src.open('rb') as rc:
            with gzip.open(dst.resolve().as_posix(), 'wb', compresslevel=9) as wc:
                while line := rc.read(1 << 24):
                    wc.write(line)

    @classmethod
    async def decompress_file(cls, src: Path, dst: Path) -> None:
        with gzip.open(src.resolve().as_posix(), 'rb') as rc:
            with dst.open('wb') as wc:
                while line := rc.read(1 << 24):
                    wc.write(line)


class Compressor:
    T_NONE = 0b0000
    T_GZIP = 0b0001
    compressors = {
        T_NONE: DummyCompressor,
        T_GZIP: GzipCompressor,
    }

    @classmethod
    async def compress(cls, buff: bytes, compression: int = None) -> (bytes, int):
        try:
            if compression is None:
                compression = await cls.propose_compression(buff)

            buff = await cls.compressors[compression].compress(buff)
            return buff, compression

        except (KeyError, ValueError, TypeError) as err:
            raise ValueError(f'Failed to compress data: {str(err)}')

    @classmethod
    async def decompress(cls, buff: bytes, compression: int) -> bytes:
        try:
            return await cls.compressors[compression].decompress(buff)
        except (KeyError, ValueError, TypeError) as err:
            raise ValueError(f'Failed to decompress data: {str(err)}')

    @classmethod
    async def compress_file(cls, src: Path, dst: Path, compression: int = None) -> int:
        try:
            if compression is None:
                compression = await cls.propose_compression(src)

            await cls.compressors[compression].compress_file(src, dst)
            return compression
        except (KeyError, ValueError, TypeError) as err:
            raise ValueError(f'Failed to compress file: {str(err)}')

    @classmethod
    async def decompress_file(cls, src: Path, dst: Path, compression: int) -> None:
        try:
            await cls.compressors[compression].decompress_file(src, dst)
        except (KeyError, ValueError, TypeError) as err:
            raise ValueError(f'Failed to decompress file: {str(err)}')

    @classmethod
    async def propose_compression(cls, buff: Union[bytes, Path]):
        if isinstance(buff, bytes):
            if len(buff) > 4096:
                return cls.T_GZIP
            else:
                return cls.T_NONE

        elif isinstance(buff, Path):
            s = os.path.getsize(buff.as_posix())
            if s > 4096:
                return cls.T_GZIP
            else:
                return cls.T_NONE

        else:
            raise TypeError('Unsupported buffer type')



from pathlib import Path
from typing import Any, AsyncIterable, Iterable, Type, Union

import ujson

from cats.errors import ProtocolError

__all__ = [
    'Bytes',
    'BytesGen',
    'BytesAsyncGen',
    'BytesAnyGen',
    'NULL',
    'Byte',
    'Json',
    'File',
    'T_Headers',
    'Headers',
]

Bytes = Union[bytes, bytearray, memoryview]
BytesGen = Iterable[Bytes]
BytesAsyncGen = AsyncIterable[Bytes]
BytesAnyGen = Union[BytesGen, BytesAsyncGen]

NULL = type('NULL', (object,), {})
Byte = Bytes
Json = Union[str, int, float, dict, list, bool, type(None), Type[NULL]]
File = Union[Path, str]

T_Headers = Union[dict[str, Any], 'Headers']


class Headers(dict):
    def __init__(self, *args, **kwargs):
        v = self._convert(*args, **kwargs)
        if (offset := v.get('offset', None)) and (not isinstance(offset, int) or offset < 0):
            raise ProtocolError('Invalid offset header')
        super().__init__(v)

    @classmethod
    def _key(cls, key: str) -> str:
        return key.replace(' ', '-').title()

    def __getitem__(self, item):
        return super().__getitem__(self._key(item))

    def __setitem__(self, key, value):
        return super().__setitem__(self._key(key), value)

    def __delitem__(self, key):
        return super().__delitem__(self._key(key))

    def __contains__(self, item):
        return super().__contains__(self._key(item))

    @classmethod
    def _convert(cls, *args, **kwargs):
        return {cls._key(k): v for k, v in dict(*args, **kwargs).items() if isinstance(k, str)}

    def update(self, *args, **kwargs) -> None:
        super().update(self._convert(*args, **kwargs))

    def encode(self) -> bytes:
        return ujson.encode(self, ensure_ascii=False,
                            escape_forward_slashes=False,
                            encode_html_chars=True).encode('utf-8')

    @classmethod
    def decode(cls, headers: Bytes) -> 'Headers':
        return cls(ujson.decode(headers.decode('utf-8') or '{}') or {})



from collections import defaultdict
from dataclasses import dataclass
from types import GeneratorType
from typing import Any, Awaitable, DefaultDict, Optional, Type, Union

import ujson

from cats.codecs import T_FILE, T_JSON
from cats.identity import Identity, IdentityChild
from cats.plugins import BaseSerializer, QuerySet
from cats.server.action import Action, BaseAction, InputAction
from cats.types import Headers, Json

__all__ = [
    'HandlerItem',
    'Api',
    'Handler',
]

T_Headers = Union[dict[str, Any], Headers]


@dataclass
class HandlerItem:
    id: int
    name: str
    handler: Type['Handler']
    version: Optional[int] = None
    end_version: Optional[int] = None


class Api:
    def __init__(self):
        self._handlers: DefaultDict[int, list[HandlerItem]] = defaultdict(list)

    def register(self, handler: HandlerItem):
        if handler in self._handlers[handler.id]:
            return

        assert handler.version is None or handler.end_version is None or handler.version <= handler.end_version, \
            f'Invalid version range for handler {handler.id}: [{handler.version}..{handler.end_version}]'

        if handler.version is not None or handler.end_version is not None:
            assert handler.version is not None, f'Initial version is not provided for {handler}'

            try:
                last_handler = self._handlers[handler.id][-1]
                assert last_handler.version is not None or last_handler.end_version is not None, \
                    f'Attempted to add versioned {handler} to wildcard'

                if last_handler.end_version is not None:
                    assert last_handler.end_version < handler.version, \
                        f'Handler {handler} overlap {last_handler} version'
                else:
                    assert last_handler.version < handler.version, \
                        f'Handler {handler} overlap {last_handler} version'
                    last_handler.end_version = handler.version - 1
            except IndexError:
                pass
        self._handlers[handler.id].append(handler)

    @property
    def handlers(self):
        return self._handlers

    def update(self, app: 'Api'):
        self._handlers.update(app.handlers)

    def compute(self) -> dict[int, Union[list[HandlerItem], HandlerItem]]:
        result: dict[int, Union[list[HandlerItem], HandlerItem]] = {}
        for handler_id, handler_list in self._handlers.items():
            if not handler_list:
                continue

            elif len(handler_list) == 1 and handler_list[0].version is None and handler_list[0].end_version is None:
                result[handler_id] = handler_list[0]
            else:
                result[handler_id] = handler_list

        return result


class Handler:
    handler_id: int

    Loader: Optional[Type[BaseSerializer]] = None
    Dumper: Optional[Type[BaseSerializer]] = None

    data_type: Optional[Union[int, tuple[int]]] = None
    min_data_len: Optional[int] = None
    max_data_len: Optional[int] = None
    block_models: Optional[tuple[str]] = None
    require_models: Optional[tuple[str]] = None
    require_auth: Optional[bool] = None
    min_file_size: Optional[int] = None
    max_file_size: Optional[int] = None
    min_file_total_size: Optional[int] = None
    max_file_total_size: Optional[int] = None
    min_file_amount: Optional[int] = None
    max_file_amount: Optional[int] = None

    def __init__(self, action: Action):
        self.action = action

    # noinspection PyShadowingBuiltins
    def __init_subclass__(cls, /, api: Api = None, id: int = None,
                          name: str = None, version: int = None, end_version: int = None):
        # abstract
        if api is None:
            return

        assert id is not None

        assert cls.Loader is None or (isinstance(cls.Loader, type) and issubclass(cls.Loader, BaseSerializer)), \
            'Handler Loader must be subclass of rest_framework.serializers.BaseSerializer'
        assert cls.Dumper is None or (isinstance(cls.Dumper, type) and issubclass(cls.Dumper, BaseSerializer)), \
            'Handler Dumper must be subclass of rest_framework.serializers.BaseSerializer'

        assert not (cls.require_auth is False and cls.require_models is not None), \
            f'{cls!s}.require_auth is False and {cls!s}.require_models is not None'

        api.register(HandlerItem(id, name, cls, version, end_version))
        cls.handler_id = id

    async def __call__(self) -> Optional[BaseAction]:
        await self.prepare()
        return await self.handle()

    async def prepare(self) -> None:
        """Called before handler() method"""
        try:
            self._check_before_recv()
        except self.action.conn.STREAM_ERRORS:
            raise
        except BaseException:
            await self.action.dump_data(self.action.data_len)
            raise
        else:
            await self.action.recv_data()
            self._check_after_recv()

    async def handle(self):
        raise NotImplementedError

    async def json_load(self, *, many: bool = False) -> Json:
        if self.action.data_type != T_JSON:
            raise TypeError('Unsupported data type. Expected JSON')

        data = self.action.data
        if self.Loader is not None:
            if many is None:
                many = isinstance(data, list)
            form = self.Loader(data=data, many=many)
            form.is_valid(raise_exception=True)
            return form.validated_data
        else:
            return data

    async def json_dump(self, data, *, headers: T_Headers = None,
                        status: int = 200, many: bool = None) -> Action:
        if many is None:
            many = isinstance(data, (list, tuple, set, QuerySet, GeneratorType))

        if self.Dumper is not None:
            data = self.Dumper(data, many=many).data
        else:
            ujson.encode(data)

        return Action(data=data, headers=headers, status=status)

    @property
    def identity(self) -> Optional[Union[Identity, IdentityChild]]:
        return self.action.conn.identity

    def ask(self, data=None, data_type: int = None, compression: int = None, *,
            headers: T_Headers = None, status: int = 200,
            bypass_limit=False, bypass_count=False, timeout=None) -> Awaitable['InputAction']:
        return self.action.ask(data, data_type=data_type, compression=compression,
                               headers=headers, status=status,
                               bypass_limit=bypass_limit, bypass_count=bypass_count, timeout=timeout)

    def _check_before_recv(self):
        self._check_data_type()
        self._check_data_len()
        self._check_models()
        self._check_auth()
        self._check_file_size()
        self._check_file_total_size()
        self._check_file_amount()

    def _check_after_recv(self):
        self._check_data_len()

    def _check_data_type(self):
        if (types := self.data_type) is None:
            return
        if isinstance(types, int):
            types = (types,)
        if self.action.data_type not in types:
            raise ValueError('Received payload type is not acceptable')

    def _check_data_len(self):
        if (x := self.action.data_len) is None:
            return
        min_len = self.min_data_len
        max_len = self.max_data_len
        if min_len is not None and x < min_len:
            raise ValueError(f'Received payload data size is less than allowed [{min_len}]')
        if max_len is not None and max_len < x:
            raise ValueError(f'Received payload data size is more than allowed [{max_len}]')

    def _check_models(self):
        model = self.identity.model_name if self.identity else None
        if self.block_models is not None and model in self.block_models:
            raise ValueError(f'Model {model} is forbidden')
        if self.require_models is not None and model not in self.require_models:
            raise ValueError(f'Model {model} is required')

    def _check_auth(self):
        if self.require_auth is None:
            return
        if self.require_auth ^ self.action.conn.signed_in():
            reason = 'required' if self.require_auth else 'forbidden'
            raise ValueError(f'Authentication is {reason}')

    def _check_file_size(self):
        if self.action.data_type != T_FILE:
            return
        for file in self.action.headers.get('Files', []):
            x = int(file['size'])
            min_size = self.min_file_size
            max_size = self.max_file_size
            if min_size is not None and x < min_size:
                raise ValueError(f'Received File[n].size is less than allowed [{min_size}]')
            if max_size is not None and max_size < x:
                raise ValueError(f'Received File[n].size is more than allowed [{max_size}]')

    def _check_file_total_size(self):
        if self.action.data_type != T_FILE:
            return

        if self.action.data_len is not None:
            x = self.action.data_len
        else:
            x = sum(file['size'] for file in self.action.headers.get('Files', []))

        min_size = self.min_file_total_size
        max_size = self.max_file_total_size
        if min_size is not None and x < min_size:
            raise ValueError(f'Received ∑ File[n].size is less than allowed [{min_size}]')
        if max_size is not None and max_size < x:
            raise ValueError(f'Received ∑ File[n].size is more than allowed [{max_size}]')

    def _check_file_amount(self):
        if self.action.data_type != T_FILE:
            return

        min_len = self.min_file_amount
        max_len = self.max_file_amount
        x = len(self.action.headers.get('Files', []))
        if min_len is not None and x < min_len:
            raise ValueError(f'Received File amount is less than allowed [{min_len}]')
        if max_len is not None and max_len < x:
            raise ValueError(f'Received File amount is more than allowed [{max_len}]')



import asyncio
from asyncio import Future
from contextlib import asynccontextmanager, contextmanager
from functools import partial
from logging import getLogger
from random import randint
from typing import Any, Iterable, Optional

from sentry_sdk import Scope, add_breadcrumb, capture_exception
from tornado.iostream import IOStream, StreamClosedError

from cats.errors import HandshakeError, ProtocolError
from cats.identity import Identity
from cats.server.action import *
from cats.server.handlers import HandlerItem
from cats.types import BytesAnyGen

__all__ = [
    'Connection',
]

logging = getLogger('CATS.conn')


class Connection:
    MAX_PLAIN_DATA_SIZE = 1 << 24
    STREAM_ERRORS = (asyncio.TimeoutError, asyncio.CancelledError, asyncio.InvalidStateError, StreamClosedError)
    IGNORE_ERRORS = (*STREAM_ERRORS, HandshakeError, ProtocolError, KeyboardInterrupt)

    __slots__ = (
        '_closed', 'stream', 'address', 'api_version', '_app', 'scope', 'download_speed',
        '_identity', '_credentials', 'loop', 'input_deq', '_idle_timer', 'message_pool', 'is_sending',
        'debug', 'recv_fut',
    )

    def __init__(self, stream: IOStream, address: tuple[str, int], api_version: int, app, debug=False):
        logging.debug(f'New connection established: {address}')
        self._closed = False
        self.stream = stream
        self.address: tuple[str, int] = address
        self.api_version = api_version
        self._app = app
        self.scope = Scope()
        self._identity: Optional[Identity] = None
        self._credentials = None
        self.loop = asyncio.get_event_loop()
        self.input_deq: dict[int, Input] = {}
        self._idle_timer: Optional[asyncio.Future] = None
        self.message_pool: list[int] = []
        self.is_sending = False
        self.download_speed = 0
        self.debug = debug
        self.recv_fut: Optional[asyncio.Future] = None

    @property
    def host(self) -> str:
        return self.address[0]

    @property
    def port(self) -> int:
        return self.address[1]

    @property
    def is_open(self):
        return not self._closed and not self.stream.closed()

    @property
    def app(self):
        return self._app

    async def init(self):
        logging.debug(f'{self} initialized')

    async def start(self, ping=False):
        if ping:
            self.loop.create_task(self.ping())

        while self.is_open:
            if self.recv_fut is not None:
                await self.recv_fut

            self.recv_fut = Future()
            task = self.loop.create_task(self.tick())
            task.add_done_callback(self.on_tick_done)

    @contextmanager
    def preserve_message_id(self, message_id: int):
        if message_id in self.message_pool:
            raise ProtocolError('Provided message_id already in use')

        try:
            self.message_pool.append(message_id)
            yield
        finally:
            try:
                self.message_pool.remove(message_id)
            except (KeyError, ValueError):
                pass

    async def recv(self):
        self.reset_idle_timer()
        action_type_id = await self.stream.read_bytes(1)
        action_class = BaseAction.get_class_by_type_id(action_type_id)
        if action_class is None:
            raise ProtocolError(f'Received unknown Action Type ID [{action_type_id.hex()}]')

        action = await action_class.init(self)
        if hasattr(action, 'recv_data'):
            await action.recv_data()
        return action

    async def tick(self):
        self.reset_idle_timer()
        action_type_id = await self.stream.read_bytes(1)
        action_class = BaseAction.get_class_by_type_id(action_type_id)
        if action_class is None:
            raise ProtocolError(f'Received unknown Action Type ID [{action_type_id.hex()}]')

        try:
            action = await action_class.init(self)
            await action.handle()
        except self.IGNORE_ERRORS as err:
            self.close(err)
        except Exception as err:
            capture_exception(err, scope=self.scope)
            self.close(err)

    def on_tick_done(self, task):
        exc = None
        try:
            exc = task.exception()
        except (asyncio.CancelledError, asyncio.InvalidStateError) as err:
            exc = err
        if exc:
            self.close(exc)

    def dispatch(self, handler_id):
        handlers = self.app.get_handlers_by_id(handler_id)
        if isinstance(handlers, HandlerItem):
            return handlers.handler

        elif isinstance(handlers, list):
            for item in handlers:
                item: HandlerItem
                end_version = self.api_version if item.end_version is None else item.end_version
                if item.version <= self.api_version <= end_version:
                    return item.handler

        raise ProtocolError(f'Handler with id {handler_id} not found')

    async def ping(self):
        if self._app.idle_timeout is None:
            return
        wait = max(0.1, round(self._app.idle_timeout / 2, 2))
        pong = PingAction()
        while self.is_open:
            await asyncio.sleep(wait)
            await pong.send(self)

    async def set_download_speed(self, speed=0):
        await self.stream.write(b'\x05')
        await self.stream.write(speed.to_bytes(4, 'big', signed=False))

    async def send(self, handler_id: int, data=None, message_id=None, compression=None, *,
                   headers=None, status=None):
        action = Action(data=data, headers=headers, status=status,
                        message_id=self._get_free_message_id() if message_id is None else message_id,
                        handler_id=handler_id, compression=compression)
        await action.send(self)

    async def send_stream(self, handler_id: int, data: BytesAnyGen, data_type: int,
                          message_id=None, compression=None, *,
                          headers=None, status=None):
        action = StreamAction(data=data, headers=headers, status=status,
                              message_id=self._get_free_message_id() if message_id is None else message_id,
                              handler_id=handler_id, data_type=data_type, compression=compression)
        await action.send(self)

    def attach_to_channel(self, channel: str):
        self.app.attach_conn_to_channel(self, channel=channel)

    def detach_from_channel(self, channel: str):
        self.app.detach_conn_from_channel(self, channel=channel)

    @property
    def conns_with_same_identity(self) -> Iterable['Connection']:
        return self.app.channel(f'model_{self._identity.model_name}:{self._identity.id}')

    @property
    def conns_with_same_model(self) -> Iterable['Connection']:
        return self.app.channel(f'model_{self._identity.model_name}')

    @property
    def identity(self) -> Optional[Identity]:
        return self._identity

    @property
    def credentials(self) -> Optional[Any]:
        return self._credentials

    @property
    def identity_scope_user(self):
        if not self.signed_in():
            return {'ip': self.host}

        return self.identity.sentry_scope

    def signed_in(self) -> bool:
        return self._identity is not None

    def sign_in(self, identity: Identity, credentials=None):
        self._identity: Optional[Identity] = identity
        self._credentials = credentials

        model_group = f'model_{identity.model_name}'
        auth_group = f'{model_group}:{identity.id}'
        self.attach_to_channel(model_group)
        self.attach_to_channel(auth_group)

        self.scope.set_user(self.identity_scope_user)
        add_breadcrumb(message='Sign in', data={
            'id': identity.id,
            'model': identity.__class__.__name__,
            'instance': repr(identity),
        })

        logging.debug(f'Signed in as {identity.__class__.__name__} <{self.host}:{self.port}>')

    def sign_out(self):
        logging.debug(f'Signed out from {self.identity.__class__.__name__} <{self.host}:{self.port}>')
        if self.signed_in():
            model_group = f'model_{self.identity.model_name}'
            auth_group = f'{model_group}:{self._identity.id}'

            self.detach_from_channel(auth_group)
            self.detach_from_channel(model_group)

            self._identity = None
            self._credentials = None

        self.scope.set_user(self.identity_scope_user)
        add_breadcrumb(message='Sign out')

        return self

    def close(self, exc=None):
        if self._closed:
            return

        self._closed = True

        self.sign_out()
        if exc and not isinstance(exc, self.IGNORE_ERRORS):
            logging.error(exc)
            capture_exception(exc, scope=self.scope)

        if self._idle_timer is not None:
            self._idle_timer.cancel()
            self._idle_timer = None
        if self.recv_fut is not None:
            self.recv_fut.cancel()
            self.recv_fut = None
        self.stream.close(exc)

    def __str__(self) -> str:
        return f'CATS.Connection: {self.host}:{self.port} api@{self.api_version}'

    def reset_idle_timer(self):
        if self.app.idle_timeout > 0:
            if self._idle_timer is not None:
                self._idle_timer.cancel()

            self._idle_timer = self.loop.call_later(self.app.idle_timeout, partial(self.close, asyncio.TimeoutError()))

    def _get_free_message_id(self) -> int:
        while True:
            message_id = randint(17783, 35565)
            if message_id not in self.message_pool:
                break

        return message_id

    @asynccontextmanager
    async def lock_write(self):
        while self.is_sending:
            await asyncio.sleep(0.05)
        self.is_sending = True
        yield
        self.is_sending = False



O_DEFAULT = '@'
O_NATIVE = '='
O_LITTLE = '<'
O_BIG = '>'
O_NETWORK = '!'

Pad = 'x'
Bool = '?'
Char = 'c'
Int1 = 'b'
uInt1 = 'B'
Int2 = 'h'
uInt2 = 'H'
Int4 = 'l'
uInt4 = 'L'
Int8 = 'q'
uInt8 = 'Q'
Float2 = 'e'
Float4 = 'f'
Float8 = 'd'
String = 's'
PascalString = 'p'



try:
    from django.db.models import QuerySet
except ImportError:
    QuerySet = type('QuerySet', (list,), {})

try:
    from rest_framework.serializers import BaseSerializer
except ImportError:
    BaseSerializer = type('BaseSerializer', (object,), {})

__all__ = [
    'QuerySet',
    'BaseSerializer'
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cats', 'cats.server']

package_data = \
{'': ['*']}

install_requires = \
['pytz>=2021.1,<2022.0',
 'sentry-sdk>=1.1.0,<2.0.0',
 'tornado>=6.1,<7.0',
 'ujson>=4.0.2,<5.0.0']

extras_require = \
{'django': ['Django>=2.0,<3.0', 'djangorestframework>=3.12.4']}

entry_points = \
{'console_scripts': ['cats_test_client = cats.client:main',
                     'cats_test_server = cats.server:main']}

setup_kwargs = {
    'name': 'cats-python',
    'version': '0.1.0b0',
    'description': 'Cifrazia Action Transport System for Python',
    'long_description': '# CATS Python\n\n## Cifrazia Action Transport System\n\n### Requirements\n\n+ Python `^3.9`\n+ Tornado `^6.1`\n+ Sentry SDK `^1.1.0`\n+ uJSON `^4.0.2`\n+ PyTZ `^2021.1`\n\n**Extras**\n\n+ `django`\n  + Django `^2.2` (optional)\n  + Django Rest Framework `^3.12.4` (optional)\n\n### Installation\n\n**Install via Poetry**\n\n```shell\npoetry add cats-python\n```\n\nWith `Django` support:\n\n```shell\npoetry add cats-python[django]\n```\n',
    'author': 'Bogdan Parfenov',
    'author_email': 'adam.brian.bright@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/AdamBrianBright/cats-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import asyncio
from dataclasses import dataclass
from functools import partial
from io import BytesIO
from logging import getLogger
from pathlib import Path
from struct import Struct
from time import time_ns
from typing import NamedTuple, Type

import math
from sentry_sdk import capture_exception

from cats.codecs import Codec, T_FILE, T_JSON
from cats.compression import Compressor
from cats.errors import MalformedDataError, ProtocolError
from cats.struct import O_NETWORK, uInt1, uInt2, uInt4, uInt8
from cats.types import Headers
from cats.utils import Delay, bytes2hex, format_amount, int2hex, tmp_file

__all__ = [
    'Input',
    'BaseAction',
    'BasicAction',
    'Action',
    'StreamAction',
    'InputAction',
    'DownloadSpeedAction',
    'CancelInputAction',
    'PingAction',
]

logging = getLogger('CATS.conn')
debug = logging.debug

MAX_IN_MEMORY = 1 << 24
MAX_CHUNK_READ = 1 << 20
PROPOSAL_PLACEHOLDER = bytes(5000)

to_uint = partial(int.to_bytes, byteorder='big', signed=False)
as_uint = partial(int.from_bytes, byteorder='big', signed=False)


class Input:
    __slots__ = ('future', 'conn', 'message_id', 'bypass_count', 'timer')

    def __init__(self,
                 future,
                 conn,
                 message_id,
                 bypass_count=False,
                 timeout=None):
        self.timer = None
        self.future = future
        self.conn = conn
        self.message_id = message_id
        self.bypass_count = bypass_count
        if timeout:
            self.timer = asyncio.get_event_loop().call_later(timeout, self.cancel)

    def cancel(self):
        if self.timer is not None:
            self.timer.cancel()
            self.timer = None
        if not self.future.done():
            self.future.cancel()
        self.conn.input_deq.pop(self.message_id, None)


class BaseAction(dict):
    __slots__ = ('data', 'headers', 'message_id', 'conn')
    __registry__ = {}
    HEADER_SEPARATOR = b'\x00\x00'

    type_id: bytes
    head_struct: Struct
    head_tuple: Type[NamedTuple]

    def __init__(self, data=None, *, headers=None, status=None,
                 message_id=None):
        assert type(self) is not BaseAction, 'Creation of BaseAction instances are prohibited'

        if headers is not None and not isinstance(headers, dict):
            raise MalformedDataError('Invalid Headers provided')

        self.data = data
        self.headers = Headers(headers or {})
        self.status = status or self.status
        self.message_id = message_id
        self.conn = None
        super().__init__()

    def __init_subclass__(cls, *, abstract=False):
        if abstract:
            return
        assert cls.type_id not in cls.__registry__, f'ActionType with ID {cls.type_id} already assigned'
        cls.__registry__[cls.type_id] = cls

    @property
    def status(self):
        return self.headers.get('Status', 200)

    @status.setter
    def status(self, value=None):
        if value is None:
            value = 200
        elif not isinstance(value, int):
            raise MalformedDataError('Invalid status type')
        self.headers['Status'] = value

    @status.deleter
    def status(self):
        self.headers['Status'] = 200

    @property
    def offset(self) -> int:
        return self.headers.get('Offset', 0)

    @offset.setter
    def offset(self, value: int = None):
        if value is None:
            value = 0
        elif not isinstance(value, int):
            raise MalformedDataError('Invalid offset type')
        self.headers['Offset'] = value

    @offset.deleter
    def offset(self):
        self.headers['Offset'] = 0

    @classmethod
    def get_class_by_type_id(cls, type_id):
        return cls.__registry__.get(type_id)

    async def ask(self, data=None, data_type=None, compression=None, *,
                  headers=None, status=None,
                  bypass_limit=False, bypass_count=False,
                  timeout=None):
        if not self.conn:
            raise ValueError('Connection is not set')
        fut = asyncio.Future()
        timeout = self.conn.app.input_timeout if timeout is None else timeout

        if not bypass_limit:
            amount = sum(1 for i in self.conn.input_deq.values() if not i.bypass_count)
            if amount > self.conn.app.INPUT_LIMIT:
                k = min(self.conn.input_deq.keys())
                self.conn.input_deq[k].cancel()

        inp = Input(fut, self.conn, self.message_id, bypass_count, timeout)
        if self.message_id in self.conn.input_deq:
            raise ProtocolError(f'Input query with MID {self.message_id} already exists')

        self.conn.input_deq[self.message_id] = inp
        action = InputAction(data, headers=headers, status=status, message_id=self.message_id,
                             data_type=data_type, compression=compression)
        await action.send(self.conn)
        return await fut

    @classmethod
    async def init(cls, conn):
        raise NotImplementedError

    @classmethod
    async def _recv_head(cls, conn):
        raise NotImplementedError

    async def dump_data(self, size: int) -> None:
        while size > 0:
            size -= len(await self.conn.stream.read_bytes(min(size, MAX_CHUNK_READ), partial=True))
        if fut := self.conn.recv_fut:
            fut.set_result(None)

    async def handle(self):
        raise NotImplementedError

    async def send(self, conn):
        raise NotImplementedError


class BasicAction(BaseAction, abstract=True):
    __slots__ = ('data_len', 'data_type', 'compression')

    def __init__(self, data=None, *, headers=None, status=None,
                 message_id=None, data_len=None, data_type=None, compression=None):
        self.data_len = data_len
        self.data_type = data_type
        self.compression = compression
        super().__init__(data, headers=headers, status=status, message_id=message_id)

    @classmethod
    async def init(cls, conn):
        raise NotImplementedError

    @classmethod
    async def _recv_head(cls, conn):
        raise NotImplementedError

    async def handle(self):
        raise NotImplementedError

    async def send(self, conn):
        raise NotImplementedError

    async def recv_data(self):
        if self.data_len > MAX_IN_MEMORY:
            if self.data_type != T_FILE:
                raise ProtocolError(f'Attempted to send message larger than '
                                    f'{format_amount(MAX_IN_MEMORY)}')
            await self._recv_large_data()
        else:
            await self._recv_small_data()

        if fut := self.conn.recv_fut:
            fut.set_result(None)

    async def _recv_small_data(self):
        left = self.data_len
        buff = bytearray()
        while left > 0:
            chunk, left = await self._recv_chunk(left)
            buff.extend(chunk)

        buff = await Compressor.decompress(buff, compression=self.compression)
        self.data = await Codec.decode(buff, self.data_type, self.headers)
        if self.data_type == T_JSON:
            debug(f'[RECV {self.conn.address}] [{int2hex(self.message_id):<4}] <- {self.data}')

    async def _recv_large_data(self):
        left = self.data_len
        src, dst = tmp_file(), tmp_file()

        try:
            with src.open('wb') as fh:
                while left > 0:
                    chunk, left = await self._recv_chunk(left)
                    fh.write(chunk)

            await Compressor.decompress_file(src, dst, compression=self.compression)
            self.data = await Codec.decode(dst, self.data_type, self.headers)
        except Exception:
            dst.unlink(missing_ok=True)
            raise
        finally:
            src.unlink(missing_ok=True)

    async def _recv_chunk(self, left):
        chunk = await self.conn.stream.read_bytes(min(left, MAX_CHUNK_READ), partial=True)
        debug(f'[RECV {self.conn.address}] [{int2hex(self.message_id):<4}] <- {bytes2hex(chunk[:64])}...')
        self.conn.reset_idle_timer()
        left -= len(chunk)
        return chunk, left

    async def _encode(self):
        data, data_type = await Codec.encode(self.data, self.headers, self.offset)
        if isinstance(data, Path):
            data, buff = tmp_file(), data
            compression = await Compressor.compress_file(buff, data, self.compression)
            data_len = data.stat().st_size
        else:
            data, compression = await Compressor.compress(data, self.compression)
            data_len = len(data)
        return data, data_len, data_type, compression

    async def _write_data_to_stream(self, conn, data, data_len, data_type, compression):
        if isinstance(data, Path):
            fh = data.open('rb')
            left = data.stat().st_size
        else:
            fh = BytesIO(data)
            left = len(data)

        try:
            max_chunk_size = min(MAX_IN_MEMORY, conn.download_speed) or MAX_IN_MEMORY
            with Delay(conn.download_speed) as delay:
                while left > 0:
                    await delay()
                    size = min(left, max_chunk_size)
                    chunk = fh.read(size)
                    left -= size
                    conn.reset_idle_timer()
                    debug(f'[SEND {conn.address}] [{int2hex(self.message_id):<4}] -> {bytes2hex(chunk[:64])}...')
                    await conn.stream.write(chunk)
        finally:
            fh.close()
            if data_type == T_JSON and isinstance(data, bytes):
                debug(f'[SEND {conn.address}] [{int2hex(self.message_id):<4}] -> {data.decode("utf-8")}')


class Action(BasicAction):
    __slots__ = ('handler_id', 'send_time')

    type_id = b'\x00'
    head_struct = Struct(f'{O_NETWORK}2{uInt2}{uInt8}2{uInt1}{uInt4}')

    @dataclass
    class Head:
        handler_id: int
        message_id: int
        send_time: int
        data_type: int
        compression: int
        data_len: int

    def __init__(self, data=None, *, headers=None, status=None, message_id=None,
                 handler_id=None, data_len=None, data_type=None, compression=None,
                 send_time=None):
        assert compression is None or isinstance(compression, int), 'Invalid compression type'
        assert data_type is None or isinstance(data_type, int), 'Invalid data type provided'

        self.handler_id = handler_id
        self.send_time = send_time or (time_ns() >> 6)
        super().__init__(data, headers=headers, status=status, message_id=message_id,
                         data_len=data_len, data_type=data_type, compression=compression)

    @classmethod
    async def init(cls, conn):
        conn.reset_idle_timer()
        head = await cls._recv_head(conn)

        headers = await conn.stream.read_until(cls.HEADER_SEPARATOR, head.data_len)
        head.data_len -= len(headers)
        headers = Headers.decode(headers[:-2])
        debug(f'[RECV {conn.address}] [{int2hex(head.message_id):<4}] <- HEADERS {headers}')
        conn.reset_idle_timer()

        inst = cls(**vars(head), headers=headers)
        inst.conn = conn
        return inst

    async def handle(self):
        with self.conn.preserve_message_id(self.message_id):
            handler = self.conn.dispatch(self.handler_id)
            fn = handler(self)
            for middleware in self.conn.app.middleware:
                fn = partial(middleware, fn)

            self.conn.message_pool.append(self.message_id)

            try:
                result = await asyncio.shield(fn())
                if result is not None:
                    if not isinstance(result, Action):
                        raise ProtocolError('Returned invalid response')

                    result.handler_id = self.handler_id
                    result.message_id = self.message_id
                    result.offset = self.offset
                    await result.send(self.conn)
            except self.conn.IGNORE_ERRORS:
                raise
            except Exception as err:
                capture_exception(err, scope=self.conn.scope)

    @classmethod
    async def _recv_head(cls, conn):
        conn.reset_idle_timer()
        buff = await conn.stream.read_bytes(cls.head_struct.size)
        head = cls.Head(*cls.head_struct.unpack(buff))
        debug(f'[RECV {conn.address}] Request  '
              f'H: {int2hex(head.handler_id):<4} '
              f'M: {int2hex(head.message_id):<4} '
              f'L: {format_amount(head.data_len):<8}'
              f'T: {Codec.codecs[head.data_type].type_name:<8} '
              f'C: {Compressor.compressors[head.compression].type_name:<8}')
        return head

    async def send(self, conn):
        data, data_len, data_type, compression = await self._encode()

        try:
            message_headers = self.headers.encode() + self.HEADER_SEPARATOR

            _data_len = data_len + len(message_headers)
            header = self.type_id + self.head_struct.pack(
                self.handler_id,
                self.message_id,
                time_ns() >> 6,
                data_type,
                compression,
                _data_len
            ) + message_headers

            async with conn.lock_write():
                conn.reset_idle_timer()
                await conn.stream.write(header)
                debug(f'[SEND {conn.address}] Response '
                      f'H: {int2hex(self.handler_id):<4} '
                      f'M: {int2hex(self.message_id):<4} '
                      f'L: {format_amount(_data_len):<8}'
                      f'T: {Codec.codecs[data_type].type_name:<8} '
                      f'C: {Compressor.compressors[compression].type_name:<8} ')
                debug(f'[SEND {conn.address}] [{int2hex(self.message_id):<4}] -> HEADERS {self.headers}')
                await self._write_data_to_stream(conn, data, data_len, data_type, compression)
        finally:
            if isinstance(data, Path):
                data.unlink(missing_ok=True)


class StreamAction(Action):
    type_id = b'\x01'
    head_struct = Struct(f'{O_NETWORK}2{uInt2}{uInt8}2{uInt1}')

    @dataclass
    class Head:
        handler_id: int
        message_id: int
        send_time: int
        data_type: int
        compression: int

    def __init__(self, data=None, *, headers=None, status=None, message_id=None,
                 handler_id=None, data_type=None, compression=None,
                 send_time=None):
        super().__init__(data, headers=headers, status=status, message_id=message_id,
                         handler_id=handler_id, data_type=data_type, compression=compression,
                         send_time=send_time)

    @classmethod
    async def init(cls, conn):
        head = await cls._recv_head(conn)
        headers_size = as_uint(await conn.stream.read_bytes(4))
        headers = Headers.decode(await conn.stream.read_bytes(headers_size))
        conn.reset_idle_timer()
        debug(f'[RECV {conn.address}] [{int2hex(head.message_id):<4}] <- HEADERS {headers}')

        action = cls(**vars(head), headers=headers)
        action.conn = conn
        return action

    @classmethod
    async def _recv_head(cls, conn):
        conn.reset_idle_timer()
        buff = await conn.stream.read_bytes(cls.head_struct.size)
        head = cls.Head(*cls.head_struct.unpack(buff))
        debug(f'[RECV {conn.address}] Stream   '
              f'H: {int2hex(head.handler_id):<4} '
              f'M: {int2hex(head.message_id):<4} '
              f'T: {Codec.codecs[head.data_type].type_name:<8} '
              f'C: {Compressor.compressors[head.compression].type_name:<8} ')
        return head

    async def recv_data(self):
        data_len = 0
        buff = tmp_file()
        try:
            with buff.open('wb') as fh:
                while chunk_size := as_uint(await self.conn.stream.read_bytes(4)):
                    self.conn.reset_idle_timer()
                    if chunk_size > MAX_IN_MEMORY:
                        data_len += await self._recv_large_chunk(fh, chunk_size)
                    else:
                        data_len += await self._recv_small_chunk(fh, chunk_size)

            if data_len > self.conn.MAX_PLAIN_DATA_SIZE:
                if self.data_type != T_FILE:
                    raise ProtocolError(f'Attempted to send message larger than '
                                        f'{format_amount(self.conn.MAX_PLAIN_DATA_SIZE)}')
                decode = buff
            elif self.data_type != T_FILE:
                with buff.open('rb') as _fh:
                    decode = _fh.read()
            self.data = await Codec.decode(decode, self.data_type, self.headers)
            self.data_len = data_len
        finally:
            if fut := self.conn.recv_fut:
                fut.set_result(None)
            buff.unlink(missing_ok=True)

    async def _recv_large_chunk(self, fh, chunk_size):
        left = chunk_size
        part, dst = tmp_file(), tmp_file()
        try:
            with part.open('wb') as tmp:
                while left > 0:
                    chunk = await self.conn.stream.read_bytes(min(left, MAX_CHUNK_READ), partial=True)
                    debug(f'[RECV {self.conn.address}] [{int2hex(self.message_id):<4}] <- {bytes2hex(chunk[:64])}...')
                    left -= len(chunk)
                    tmp.write(chunk)
            await Compressor.decompress_file(part, dst, compression=self.compression)
            data_len = dst.stat().st_size
            with dst.open('rb') as tmp:
                while i := tmp.read(MAX_IN_MEMORY):
                    fh.write(i)
            return data_len
        finally:
            part.unlink(missing_ok=True)
            dst.unlink(missing_ok=True)

    async def _recv_small_chunk(self, fh, chunk_size):
        left = chunk_size
        part = bytearray()
        while left > 0:
            chunk = await self.conn.stream.read_bytes(min(left, MAX_CHUNK_READ), partial=True)
            debug(f'[RECV {self.conn.address}] [{int2hex(self.message_id):<4}] <- {bytes2hex(chunk[:64])}...')
            left -= len(chunk)
            part += chunk
        part = await Compressor.decompress(part, compression=self.compression)
        fh.write(part)
        return len(part)

    async def _encode_gen(self, conn):
        data = self.data
        compression = self.compression
        if compression is None:
            compression = await Compressor.propose_compression(PROPOSAL_PLACEHOLDER)

        assert hasattr(data, '__iter__') or hasattr(data, '__aiter__'), \
            'StreamResponse payload is not (Async)Generator[Bytes, None, None]'
        data = self._async_gen(data, conn.download_speed)
        return data, compression

    async def send(self, conn):
        data, compression = await self._encode_gen(conn)

        header = self.type_id + self.head_struct.pack(
            self.handler_id,
            self.message_id,
            time_ns() >> 6,
            self.data_type,
            compression
        )
        message_headers = self.headers.encode()

        async with conn.lock_write():
            conn.reset_idle_timer()
            await conn.stream.write(header)
            await conn.stream.write(to_uint(len(message_headers), 4))
            await conn.stream.write(message_headers)
            debug(f'[SEND {conn.address}] Stream   '
                  f'H: {int2hex(self.handler_id):<4} '
                  f'M: {int2hex(self.message_id):<4} '
                  f'T: {Codec.codecs[self.data_type].type_name:<8} '
                  f'C: {Compressor.compressors[compression].type_name:<8} ')
            debug(f'[SEND {conn.address}] [{int2hex(self.message_id):<4}] -> HEADERS {self.headers}')
            await self._write_data_to_stream(conn, data, compression=compression)

    async def _write_data_to_stream(self, conn, data, *_, compression):
        offset = self.offset
        with Delay(conn.download_speed) as delay:
            async for chunk in data:
                if offset > 0:
                    offset -= (i := min(offset, len(chunk)))
                    chunk = chunk[i:]
                if not chunk:
                    continue
                await delay()
                if not isinstance(chunk, (bytes, bytearray, memoryview)):
                    raise MalformedDataError('Provided data chunk is not binary')

                chunk, _ = await Compressor.compress(chunk, compression)
                chunk_size = len(chunk)
                if chunk_size >= 1 << 32:
                    raise ProtocolError('Provided data chunk exceeded max chunk size')

                conn.reset_idle_timer()
                await conn.stream.write(to_uint(chunk_size, 4))
                await conn.stream.write(chunk)
                debug(f'[SEND {conn.address}] [{int2hex(self.message_id):<4}] -> {bytes2hex(chunk[:64])}...')
        await conn.stream.write(b'\x00\x00\x00\x00')

    async def _async_gen(self, gen, chunk_size):
        size = max(chunk_size, MAX_IN_MEMORY) or MAX_IN_MEMORY
        if hasattr(gen, '__iter__'):
            for item in gen:
                for i in range(math.ceil(len(item) / size)):
                    yield item[i * size:]
        else:
            async for item in gen:
                for i in range(math.ceil(len(item) / size)):
                    yield item[i * size:]


class InputAction(BasicAction):
    type_id = b'\x02'
    head_struct = Struct(f'{O_NETWORK}{uInt2}2{uInt1}{uInt4}')

    @dataclass
    class Head:
        message_id: int
        data_type: int
        compression: int
        data_len: int

    @classmethod
    async def init(cls, conn):
        head = await cls._recv_head(conn)
        headers = await conn.stream.read_until(cls.HEADER_SEPARATOR, head.data_len)
        conn.reset_idle_timer()
        head.data_len -= len(headers)
        headers = Headers.decode(headers[:-2])
        debug(f'[RECV {conn.address}] [{int2hex(head.message_id):<4}] <- HEADERS {headers}')

        action = cls(**vars(head), headers=headers)
        action.conn = conn
        return action

    @classmethod
    async def _recv_head(cls, conn):
        conn.reset_idle_timer()
        buff = await conn.stream.read_bytes(cls.head_struct.size)
        head = cls.Head(*cls.head_struct.unpack(buff))
        debug(f'[RECV {conn.address}] Answer   '
              f'M: {int2hex(head.message_id):<4} '
              f'L: {format_amount(head.data_len):<8}'
              f'T: {Codec.codecs[head.data_type].type_name:<8} '
              f'C: {Compressor.compressors[head.compression].type_name:<8} ')
        return head

    async def handle(self):
        await self.recv_data()
        if inp := self.conn.input_deq.get(self.message_id, None):
            inp.future.set_result(self)
            inp.cancel()
        else:
            raise ProtocolError('Received answer but input does`t exists')

    async def send(self, conn):
        data, data_len, data_type, compression = await self._encode()
        try:
            message_headers = self.headers.encode() + self.HEADER_SEPARATOR
            _data_len = data_len + len(message_headers)
            header = self.type_id + self.head_struct.pack(
                self.message_id,
                data_type,
                compression,
                _data_len
            ) + message_headers

            async with conn.lock_write():
                conn.reset_idle_timer()
                await conn.stream.write(header)
                debug(f'[SEND {conn.address}] Input    '
                      f'M: {int2hex(self.message_id):<4} '
                      f'L: {format_amount(_data_len):<8}'
                      f'T: {Codec.codecs[data_type].type_name:<8} '
                      f'C: {Compressor.compressors[compression].type_name:<8} ')
                debug(f'[SEND {conn.address}] [{int2hex(self.message_id):<4}] -> HEADERS {self.headers}')
                await self._write_data_to_stream(conn, data, data_len, data_type, compression)
        finally:
            if isinstance(data, Path):
                data.unlink(missing_ok=True)

    async def reply(self, data=None, data_type=None, compression=None, *,
                    headers=None, status=None):
        action = InputAction(data, headers=headers, status=status, message_id=self.message_id,
                             data_type=data_type, compression=compression)
        action.offset = self.offset
        await action.send(self.conn)

    async def cancel(self):
        res = CancelInputAction(self.message_id)
        await res.send(self.conn)


class DownloadSpeedAction(BaseAction):
    type_id = b'\x05'
    head_struct = Struct(f'{O_NETWORK}{uInt4}')

    @dataclass
    class Head:
        speed: int

    def __init__(self, speed):
        super().__init__()
        self.speed = speed

    @classmethod
    async def init(cls, conn):
        conn.reset_idle_timer()
        buff = await conn.stream.read_bytes(cls.head_struct.size)
        head = cls.Head(*cls.head_struct.unpack(buff))
        debug(f'[RECV {conn.address}] SET Download speed: {format_amount(head.speed)}')
        action = cls(**vars(head))
        action.conn = conn
        return action

    async def handle(self):
        limit = self.speed
        if not limit or (1024 <= limit <= 33_554_432):
            self.conn.download_speed = limit
        else:
            raise ProtocolError('Unsupported download speed limit')
        await self.dump_data(0)

    async def send(self, conn):
        async with conn.lock_write():
            conn.reset_idle_timer()
            speed: int = self.data
            await conn.stream.write(self.type_id + to_uint(speed, 4))
            debug(f'[SEND {conn.address}] SET Download speed: {format_amount(speed)}')


class CancelInputAction(BaseAction):
    type_id = b'\x06'
    head_struct = Struct(f'{O_NETWORK}{uInt2}')

    @dataclass
    class Head:
        message_id: int

    @classmethod
    async def init(cls, conn):
        conn.reset_idle_timer()
        buff = await conn.stream.read_bytes(cls.head_struct.size)
        head = cls.Head(*cls.head_struct.unpack(buff))
        debug(f'[RECV {conn.address}] CANCEL Input M: {int2hex(head.message_id):<4}')
        action = cls(**vars(head))
        action.conn = conn
        return action

    async def handle(self):
        if self.message_id in self.conn.input_deq:
            self.conn.input_deq[self.message_id].cancel()
        await self.dump_data(0)

    async def send(self, conn):
        async with conn.lock_write():
            conn.reset_idle_timer()
            message_id: int = self.data
            await conn.stream.write(self.type_id + to_uint(message_id, 2))
            debug(f'[SEND {conn.address}] CANCEL Input M: {message_id}')


class PingAction(BaseAction):
    type_id = b'\xFF'
    head_struct = Struct(f'{O_NETWORK}{uInt8}')

    @dataclass
    class Head:
        send_time: int

    def __init__(self, send_time=None):
        super().__init__()
        self.recv_time = time_ns() >> 6
        self.send_time = send_time or self.recv_time

    @classmethod
    async def init(cls, conn):
        conn.reset_idle_timer()
        buff = await conn.stream.read_bytes(cls.head_struct.size)
        head = cls.Head(*cls.head_struct.unpack(buff))
        debug(f'[PING {conn.address}] {head.send_time}')

        action = cls(**vars(head))
        action.conn = conn
        return action

    async def handle(self):
        debug(f'Ping {self.data.send_time} [-] {self.data.recv_time}')
        await self.send(self.conn)
        await self.dump_data(0)

    async def send(self, conn):
        async with conn.lock_write():
            conn.reset_idle_timer()
            now = time_ns() >> 6
            await conn.stream.write(self.type_id + to_uint(now, 8))
            debug(f'[SEND {conn.address}] PONG {now}')
